/**
 * JASMINe
 * Copyright (C) 2009 Bull S.A.S.
 * Contact: jasmine@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: RuleEngineService.java 4798 2009-08-07 12:30:47Z alitokmen $
 * --------------------------------------------------------------------------
 */
package org.ow2.jasmine.monitoring.eos.rules.common;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.osgi.framework.ServiceReference;
import org.ow2.jasmine.monitoring.eos.common.osgi.OSGiManager;
import org.ow2.jasmine.rules.manager.model.IRulePackage;
import org.ow2.jasmine.rules.manager.model.textual.ITextualRulePackage;
import org.ow2.jasmine.rules.manager.model.visual.ILHSElement;
import org.ow2.jasmine.rules.manager.model.visual.IRHSElement;
import org.ow2.jasmine.rules.manager.model.visual.IRule;
import org.ow2.jasmine.rules.manager.model.visual.IVisualRulePackage;
import org.ow2.jasmine.rules.manager.model.visual.Rule;
import org.ow2.jasmine.rules.manager.model.visual.VisualRulePackage;
import org.ow2.jasmine.rules.manager.service.DroolsManagerFactory;
import org.ow2.jasmine.rules.manager.service.IDroolsManager;
import org.ow2.jasmine.rules.manager.service.IDroolsManagerFactory;
import org.ow2.jasmine.rules.manager.service.RuleEngineStatus;
import org.ow2.jasmine.rules.manager.service.RulePackageCompilationError;
import org.ow2.jasmine.rules.manager.service.RulePackageSaveReport;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Rule engine server-side service. Only one running instance of this service is
 * needed because it manage several rule engine instances à the same time. Provides
 * all the methods needed by the client side rules-related components.
 *
 * @author Julien Nicoulaud
 * @see {@link org.ow2.jasmine.rules.manager.service.IDroolsManager}
 * @see {@link org.ow2.jasmine.rules.manager.service.IDroolsManagerFactory}
 */
public class RuleEngineService implements Serializable {

	// =========================================================================
	// Variables
	// =========================================================================

	/**
	 * Serial version ID
	 */
	private static final long serialVersionUID = 1943812711958376631L;

	/**
	 * Logger
	 */
	private static Log logger = LogFactory.getLog(RuleEngineService.class);

	/**
	 * The OSGi manager
	 */
	OSGiManager osgiManager = new OSGiManager();

	/**
	 * The manager factory
	 */
	IDroolsManagerFactory managerFactory;

	/**
	 * The Drools Managers
	 */
	private Map<String,IDroolsManager> droolsManagers =
		new HashMap<String,IDroolsManager>();


	// =========================================================================
	// Public methods - client side provided API
	// =========================================================================

	// Rule engine control

	/**
	 * Start the rule engine
	 */
	public void start(final String ruleEngineName) throws Exception {

		// Retrieve the rule engine manager
		IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

		// Start the rule engine
		ruleEngineManager.start();
	}

	/**
	 * Stop the rule engine
	 */
	public void stop(final String ruleEngineName) throws Exception {

		// Retrieve the rule engine manager
		IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

		// Start the rule engine
		ruleEngineManager.stop();
	}


	// Status & Statistics

	/**
	 * Get the rule engine status
	 */
	public RuleEngineStatus getRuleEngineStatus(final String ruleEngineName)
		throws Exception {

		// Retrieve the rule engine manager
		IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

		// Build a Rule Engine Status object
		RuleEngineStatus res = new RuleEngineStatus();

		//// Set the name
		res.ruleEngineName = ruleEngineName;

		//// Set the state
		res.engineStatus = ruleEngineManager.getState().toString();

		//// Set some details
		switch(ruleEngineManager.getState()) {
			case IDLE:
				res.engineDetails = "The rule engine is available and waiting" +
									" for rules to run.";
				break;
			case RUNNING:
				res.engineDetails = "The rule engine is running. ";
				List<String> deployedPackagesNames = ruleEngineManager.getDeployedPackagesNames();
				if (deployedPackagesNames != null) {
				    if (deployedPackagesNames.size() > 1) {
	                    res.engineDetails += deployedPackagesNames.size() + " packages deployed";
				    } else {
                        res.engineDetails += deployedPackagesNames.size() + " package deployed";
				    }
				}
				break;
			case STOPPED:
				res.engineDetails = "The rule engine is stopped.";
				break;
		}

		return res;
	}

	/**
	 * Get all the rule packages associated to a given rule engine
	 *
	 * @param ruleEngineName
	 * @return
	 * @throws Exception
	 */
	public List<IRulePackage> getRulePackages(final String ruleEngineName)
		throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Get the packages
        IRulePackage[] res = ruleEngineManager.getRulePackages();

        // Return the packages
        if (res != null) {
            return Arrays.asList(res);
        }
        return new ArrayList<IRulePackage>();
	}

	/**
	 * Get a rule package identified by its unique systemName attribute
	 *
	 * @param rulePackageSystemName
	 * @param ruleEngineName
	 * @return
	 */
	public IRulePackage getRulePackage(final String rulePackageSystemName,
									   final String ruleEngineName) throws Exception {


        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Return the package
        return ruleEngineManager.getRulePackage(rulePackageSystemName);
	}

	/**
	 * Commit a rule package changes
	 *
	 * @param rulePackage The rule package to save
	 * @param ruleEngineName The target rule engine instance
	 * @return A "save" operation report to inform precisely the client side
	 * @throws Exception
	 */
	public RulePackageSaveReport saveRulePackage(final IRulePackage rulePackage,
						       					 final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Apply changes
        return ruleEngineManager.commitRulePackageChanges(rebuildPackage(rulePackage));
	}

    /**
     * Delete a rule package from the database
     *
     * @param rulePackage The rule package to save
     * @param ruleEngineName The target rule engine instance
     * @throws Exception
     */
    public void deleteRulePackage(final IRulePackage rulePackage,
                                  final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Apply changes
        ruleEngineManager.deleteRulePackage(rebuildPackage(rulePackage));
    }

    /**
     * Test if a rule package compiles well
     * @param rulePackage
     * @param ruleEngineName
     * @throws Exception
     */
    public List<RulePackageCompilationError> compileRulePackage(final IRulePackage rulePackage,
                                                                final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Run the service command and return the result
        return ruleEngineManager.checkCompilation(rebuildPackage(rulePackage));
    }

    /**
     * Test if a rule package name is available
     * @param systemName the name to check availability
     * @param ruleEngineName
     * @throws Exception
     */
    public Boolean checkNameAvailability(final String systemName,
                                         final String ruleEngineName) throws Exception {

        // Retrieve the rule engine manager
        IDroolsManager ruleEngineManager = getRuleEngineManager(ruleEngineName);

        // Run the service command and return the result
        return ruleEngineManager.checkNameAvailability(systemName);
    }


	// =========================================================================
	// Private methods
	// =========================================================================

	/**
	 * Get or create the rule engine manager associated to a rule engine instance
	 */
	private IDroolsManager getRuleEngineManager(final String ruleEngineName)
		throws Exception {

	    // If no registered manager, retrieve it
		if(!droolsManagers.containsKey(ruleEngineName)){

		    try {

    		    // Get the manager factory service reference
    		    ServiceReference managerFactoryRef =
    		        osgiManager.getBundleContext().getServiceReference(IDroolsManagerFactory.class.getName());

    		    // Get its service instance
                managerFactory = (DroolsManagerFactory) osgiManager.getBundleContext().getService(managerFactoryRef);

                // Get a manager from it and store it
                droolsManagers.put(ruleEngineName,managerFactory.getManager(ruleEngineName));

		    } catch (Exception e) {
		        throw new Exception("Could not connect to the rule engine manager",e);
		    }
		}

		// Return the manager
		return droolsManagers.get(ruleEngineName);
	}

	/**
	 * Rebuild manually a given rule package. This is a workaround for the fact
	 * that Flex builds the Java objects with its own subtypes (for instance,
	 * flex.messaging.io.ArrayCollection that extends ArrayList).
	 *
	 * @param rulePackage the package to rebuild
	 * @return the package built with no flex.* class
	 */
	private IRulePackage rebuildPackage(final IRulePackage rulePackage) {
        if (rulePackage instanceof ITextualRulePackage) {
            return rulePackage;
        }
        if (rulePackage instanceof IVisualRulePackage) {
            IVisualRulePackage convertedRP = new VisualRulePackage();
            convertedRP.setDescription(rulePackage.getDescription());
            convertedRP.setDisplayName(rulePackage.getDisplayName());
            convertedRP.setPreviousSystemName(rulePackage.getPreviousSystemName());
            convertedRP.setRuleEngineInstanceName(rulePackage.getRuleEngineInstanceName());
            convertedRP.setState(rulePackage.getState());
            convertedRP.setSystemName(rulePackage.getSystemName());

            ArrayList<String> convertedImports = new ArrayList<String>();
            for (String importStatement : ((IVisualRulePackage) rulePackage).getImports()) {
                convertedImports.add(importStatement);
            }
            convertedRP.setImports(convertedImports);

            ArrayList<IRule> convertedRules = new ArrayList<IRule>();
            for (IRule rule : ((IVisualRulePackage)rulePackage).getRules()) {
                IRule convertedRule = new Rule();
                convertedRule.setName(rule.getName());

                ArrayList<String> convertedVariables = new ArrayList<String>();
                for (String variable : rule.getVariables()) {
                    convertedImports.add(variable);
                }
                convertedRule.setVariables(convertedVariables);

                ArrayList<ILHSElement> convertedLHS = new ArrayList<ILHSElement>();
                for (ILHSElement lhsElement : rule.getLHS()) {
                    convertedLHS.add(lhsElement);
                }
                convertedRule.setLHS(convertedLHS);

                ArrayList<IRHSElement> convertedRHS = new ArrayList<IRHSElement>();
                for (IRHSElement rhsElement : rule.getRHS()) {
                    convertedRHS.add(rhsElement);
                }
                convertedRule.setRHS(convertedRHS);

                convertedRules.add(convertedRule);
            }
            convertedRP.setRules(convertedRules);

            return convertedRP;
        }
	    return null;
	}

}
